
package HostingMaster::CGI;

use strict;
use CGI;
use HostingMaster::Client;
use vars qw( @ISA );

BEGIN { 
    @ISA = qw( HostingMaster::Client ); 
}

sub new {
    my $class = ref $_[0] ? ref $_[0] : shift;
    my $self = bless { ref $_[0] ? %{$_[0]} : @_ }, $class;

    $self = $self->SUPER::new( @_ );

    $self->{'cgi'} = new CGI;

    $self->{'input'} = $self->{'output'} = $self->{'status'} = '';
    $self->{'page'} = $self->cleansubmit('do','\w+') || 'login';
    $self->{'username'} = $self->cleanparam('username','\w{1,8}');
    $self->{'secret'} = $self->cleanparam('id','[\w-]{1,99}');
    if( ! $self->{'secret'} ) {
	$self->{'password'} = 
	    $self->cleanparam('password','[\x20-\xFE]{1,99}');
	$self->{'session'} = 'begin';
    }

    warn "$0 started:", map( {" $_=$self->{$_}"} sort keys %$self ), "\n";

    return $self;
}

sub clean {
    shift if ref $_[0]; # static or method
    my ( $dirty, $re ) = @_;
    return( ( $dirty !~ /\'/ and $dirty =~ /^($re)$/ ) ? $1 : undef );
}

sub cleanparam {
    my ( $self, $name, $re ) = @_;
    return clean( $self->{'cgi'}->param($name), $re );
}

sub cleansubmit {
    my ( $self, $name, $re ) = @_;
    foreach( $self->{'cgi'}->param() ) {
	next unless /^(?:$name)_(.*)$/;
	my $result = clean( $1, $re );
	return $result;
    }
    return;
}

sub cleanmultiline {
    my ( $self, $name, $re, $maxlen ) = @_;
    my @clean;
    $maxlen ||= 99;
    foreach( split("\n", $self->{'cgi'}->param($name)) ) {
        s/^\s+//; s/\s+$//;
	my $row = clean( $_, $re );
	die "Bad input for '$name' $re: '$_'\n"
	    unless defined $row;
	die "Maximum length exceeded for '$name'\n"
	    unless $maxlen--;
	push @clean, $row;
    }
    return \@clean;
}

sub readfile {
    my ( $self, $filename ) = @_;
    my $f = "$self->{'datadir'}/$filename.html";
    open F, $f or die "$0 failed to read $f: $!\n";
    my $contents = join '', <F>;
    close F;
    return $contents;
}

sub cgi_output {
    my $self = shift;
    $self->{'status'} =~ s/\n/<br\/>\n/gs;

    if( @_ ) {
	$self->{'page'} = shift unless ref $_[0];
	my $overrides = shift;
	if( ref $overrides ) {
	    foreach( keys %$overrides ) {
		$self->{$_} = $overrides->{$_};
	    }
	}
    }

    if( $self->{'page'} ne 'login' and not $self->{'authenticated'} ) {
	eval {
	    $self->master_execute('authenticate');
	};
	$self->{'page'} = 'login'
	    unless $self->{'authenticated'};
    }

    print $self->{'cgi'}->header() unless $self->{'header_printed'}++;

    my $main_template = $self->readfile('main_template');

    $self->{'page_template'} = $self->readfile($self->{'page'});

    $self->{'main_footer'} = $self->{'main_header'} = '';
    if( $self->{'page'} ne 'login' ) {
	$self->{'main_header'} = 
	    "<input type=\"hidden\" name=\"username\" value=\"\${username}\" />" .
		"<input type=\"hidden\" name=\"id\" value=\"\${secret}\" />";
	if( $self->{'page'} ne 'menu' ) {
	    $self->{'main_footer'} = 
		"<input type=submit name=do_menu value=\"Return to Menu\" />";
	    if( $self->{'page'} ne 'domain' and 
		$self->{'domain'} and $self->{'domain'} ne 'NEW' ) {
		$self->{'main_footer'} .= 
		    " &nbsp; <input type=submit name=do_domain " .
			"value=\"Return to Domain \${domain}\" /> ";
	    }
	}
    }

    while( $main_template =~ s/\$\{(\w+)\}/$self->{$1}/g ) {};

    print $main_template;
}

sub master_execute {
    my ( $self, $cmdname ) = @_;
    my $cmd = "$self->{'cmd'}{$cmdname}"
	or die "Missing command: $cmdname\n";

    while( $cmd =~ s/\$\{(\w+)\}/$self->{$1}/g ) {}

    # warn "$0 running command $cmdname: $cmd\n";
    eval {
	$self->run( $cmd );
    };
    warn "$0 ran command $cmdname: $cmd ($@)\n";
    # warn "$0 got error: $@" if $@;

    $self->{'master_result'} = my $result = $@ || $self->{'output'};
    if( $result =~ /^Failed:/i ) {
	$self->{'page'} = 'login' if $result =~ /master password/;
	$self->{'status'} = $result;
	die "$self->{'status'}\n";
    }
    $self->{'authenticated'} = 1;
    my $report = $result;
    $report =~ s/\s+/ /gs;
    $report =~ s/^(.{56}).*?(.{20})$/$1...$2/;
    # warn "$0 command executed succesfully ($report)\n";
    return $result;
}

sub domains {
    my $self = shift;
    return split /\s+/s, $self->master_execute('domainlist');
}

sub maildirects {
    my ( $self, $single ) = @_;
    my ( %addresses, %receivers );
    my $address;
    my $cmd = $single ? 'mailaddresslist' : 'mailaddresseslist';
    foreach( split("\n", $self->master_execute($cmd)) ) {
	if( /^(.*):$/ ) {
	    $address = $1;
	} elsif( /^\s+(.*)$/ and $address ) {
	    next if $address =~ /[^\w.=+-]/;
	    $addresses{$address} ||= [];
	    next unless $1;
	    push @{$addresses{$address}}, $1;
	    push @{$receivers{$1} ||= []}, $address;
	} else {
	    die "Bad input from mailaddresslist.\n";
	}
    }
    return( ( $single? $addresses{$address} : \%addresses ), 
	    \%receivers );
}

sub page {
    my ( $self, $page, $status ) = @_;
    $self->{'page'} = $page if $page;
    $self->{'status'} = $status if $status;
    my $method = "page_$self->{'page'}";

    warn "$0 page $self->{'page'} ($self->{'status'})\n";

    eval {
	$self->$method();
	$self->cgi_output();
    };
    warn "$0 page error: $@" if $@;

    $self->page( ($self->{'page'} ne 'menu' ? 'menu' : 'login'), $@ )
	if $@ and $self->{'page'} ne 'login';

    exit 0;
}

### PAGE METHODS

sub page_login {
    my $self = shift;
    $self->{'status'} .= " Please login with a valid master username";
}

sub page_menu {
    my $self = shift;

    # a call to master is required at this point to validate login
    eval {
	$self->master_execute('authenticate');
    };

    $self->page('login',$@) if $@;

    my @domains = $self->domains();
    $self->{'domainlist'} = '';
    foreach( sort @domains ) {
	$self->{'domainlist'} .= 
	    "\t<OPTION VALUE=\"$_\">$_</OPTION>\n";
    }
    $self->{'domainactions'} =
	" <INPUT TYPE=\"submit\" " .
	    "NAME=\"do_domain\" VALUE=\"Choose\"></TD>\n</TR>\n";
}

sub page_domain {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','(?:[\w.-]{6,99}|NEW)')
	or $self->page('menu','Missing domain name');
    $self->page('new_domain') if $self->{'domain'} eq 'NEW';

    my $users = $self->master_execute('listusers');
    my ( $maildirects, $mailreceivers ) = $self->maildirects(0);

    $self->{'userlist'} = '';
    foreach my $userrow ( split("\n", $users) ) {
	my @user = split /;\s*/, $userrow;
	if( $user[1] ) {
	    my $maildir = "$self->{'domain'}/mail";
	    my $addresses = $mailreceivers->{"$maildir/$user[1]/"} || [];
	    $user[1] = @$addresses>1 ? "(".@$addresses." mails)" : 
		@$addresses ? "($addresses->[0])" : '';
	}
	$user[2] &&= "ftp:$user[2]";
	$self->{'userlist'} .=
	    "\t<OPTION VALUE=\"$user[0]\">@user</OPTION>\n";
    }

    $self->{'mailaddresslist'} = '';
    foreach my $address ( sort keys %$maildirects ) {
	my @directs = @{$maildirects->{$address}};
	my $info = @directs == 1 ? $directs[0] : ( @directs . " addresses" );
	$info =~ s/^\&//;
	$info =~ s|.*?/([^/]+)/*$|$1|;
	$info = "($info)" if $info;
	$self->{'mailaddresslist'} .=
	    "\t<OPTION VALUE=\"$address\">$address $info</OPTION>\n";
    }

    $self->{'remove_button'} = 
	'<INPUT TYPE=SUBMIT NAME=do_remove_domain VALUE="Remove Domain" />';
}

sub page_new_domain {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','(?:[\w.-]{6,99}|NEW)')
	or $self->page('menu','Missing domain name');
    return if $self->{'domain'} eq 'NEW';
    $self->master_execute('domainadd');
    $self->page('domain', 'Domain created.');
}

sub page_remove_domain {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    if( $self->cleanparam('sure','SURE') eq 'SURE' ) {
	$self->master_execute('domainremove');
	$self->page('menu', "Domain '$self->{'domain'}' removed.");
    }
}

sub page_user {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    my $user = $self->cleanparam('user','(?:[\w]{1,8}|NEW)')
	or $self->page('domain','Missing user name');
    $self->page('new_user') if $user eq 'NEW';

    $self->{'user'} = $user;

    my @rows = 
	grep /^$user;/, split "\n", $self->master_execute('listusers');
    die "Bad number of rows for $user\@$self->{'domain'}\n" if $#rows;
    chomp $rows[0];
    $rows[0] =~ /^$user\;\s*([^\;]*)\;\s*(.*)$/
	or die "Bad syntax in user list for $user\n";
    my ( $old_mail, $old_ftpdir ) = ( $1, $2 );

    $self->{'services'} =
	join ',', ( ($old_mail? 'mail': ()),
		    ($old_ftpdir? 'ftp': ()) );

    $self->{'password'} =
	$self->cleanparam('password', '[\x20-\xFE]{1,20}');

    $self->{'mail'} = '';
    if( $old_mail ) {
	my $maildir = "$self->{'domain'}/mail";
	my ( undef, $mailreceivers ) = $self->maildirects(0);
	foreach( @{$mailreceivers->{"$maildir/$old_mail/"}||[]} ) {
	    $self->{'mail'} .= "$_\@$self->{'domain'}<br/>\n"
		if $_;
	}
    }

    $self->{'ftpdir'} =
	$self->cleanparam('ftpdir', '[\/\w.-]{1,99}') || $old_ftpdir;

    if( defined $self->{'ftpdir'} and 
	$self->{'ftpdir'} ne $old_ftpdir ) {
	if( $self->{'ftpdir'} and not $self->{'password'} ) {
	    $self->{'status'} = 'Password required for FTP account';
	    return;
	}
	if( $old_ftpdir ) {
	    $self->{'services'} = 'ftp';
	    $self->master_execute('removeuser');
	    $self->{'status'} .= " FTP account '$user' removed."
		unless $self->{'ftpdir'};
	}
	if( $self->{'ftpdir'} ) {
	    $self->master_execute('ftpuseradd');
	    $self->{'status'} .= " FTP account '$user' " .
		($old_ftpdir ? "changed.": "created.");
	}
	$self->{'services'} = $old_mail ? 'mail' : '';
    }

    if( $self->{'services'} and $self->{'password'} ) {
	$self->master_execute('setpwd');
	$self->{'status'} .= 
	    " Password set for $user\@$self->{'domain'} " .
		"$self->{'services'}.";
    }

    $self->{'remove_button'} = 
	'<INPUT TYPE=SUBMIT NAME=do_remove_user VALUE="Remove User" />';
}

sub page_new_user {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    my $user = $self->cleanparam('user','(?:[\w]{1,8}|NEW)')
	or $self->page('domain','Missing user name');
    return if $user eq 'NEW';
    $self->{'user'} = $user;
    $self->{'password'} = $self->cleanparam('user','[\x20-\xFE]{1,20}');
    $self->{'mail'} = $self->cleanparam('mail','[\w.-]{0,99}');
    $self->{'ftpdir'} = $self->cleanparam('ftpdir','[\w.\/-]{0,99}');
    my $freename = $self->master_execute('usernameisfree');
    chomp $freename;
    if( $freename ne $user ) {
	$self->{'status'} = 'Username is reserved. Choose another.';
	$self->{'user'} = $freename;
	return;
    }
    if( ! $self->{'password'} ) {
	$self->{'status'} = 'Password required.';
	return;
    }
    if( ! ( $self->{'mail'} or $self->{'ftpdir'} ) ) {
	$self->{'status'} = 'Specify Email address, FTP directory or both.';
	return;
    }
    if( $self->{'mail'} ) {
	$self->master_execute('mailuseradd');
	$self->{'status'} .= ' Mailbox created.';
    }
    if( $self->{'ftpdir'} ) {
	$self->{'ftpdir'} = "/$self->{'ftpdir'}"
	    unless $self->{'ftpdir'} =~ m|^/|;
	$self->master_execute('ftpuseradd');
	$self->{'status'} .= ' FTP account created.';
    }
    $self->page('user');
}

sub page_remove_user {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    $self->{'user'} = $self->cleanparam('user','(?:[\w]{1,8}|NEW)')
	or $self->page('domain','Missing user name');
    if( $self->cleanparam('sure','SURE') eq 'SURE' ) {
	$self->{'services'} = 'mail,ftp';
	$self->master_execute('removeuser');
	$self->page('domain', "User '$self->{'user'}' removed.");
    }
}

sub page_mail {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    my $mail = $self->cleanparam('mail','(?:[\w.-]{1,99}|NEW)');
    $self->page('new_mail') if $mail eq 'NEW';

    $self->{'mail'} = $mail;

    my $homedir = (getpwnam($self->{'username'}))[7] or
	die "Missing home directory of '$self->{'username'}'\n";
    my $maildir = "$self->{'domain'}/mail";
    my $mailto = $self->cleanmultiline('mailto','[\/\w\@.-]{0,999}');

    my ( $old_mailto ) = $self->maildirects(1);
    my %old_mailto_has = map { s/^&//; $_ => 1 } @$old_mailto;
    my %cur_mailto = map { $_ => 1 } keys %old_mailto_has;

    if( @$mailto ) {
	my ( %new_mailto, @removed, @added );
	my $err;
	foreach( @$mailto ) {
	    s/^\s+//; s/\s+$//;
	    next unless $_;
	    if( /^[\w.=+-]+\@[\w.-]+?\.\w{2,6}$/ ) {
		# correct mail address
	    } elsif( /^\/*([\/\w.-]+?)\/*$/ ) {
		if( -e "$homedir/$maildir/$1" ) {
		    $_ = "$maildir/$1";
		    $_ .= '/' if -d "$homedir/$_";
		} elsif( -e "$homedir/$1" ) {
		    $_ = "$1";
		    $_ .= '/' if -d "$homedir/$_";
		} elsif( /^\w{1,20}$/ and getpwnam($_) ) {
		    # correct unix username
		} else {
		    $err .= " Redirect to nonexistent mailbox '$homedir/$_'\n";
		}
	    } else {
		$err .= " Bad input. One email address or mailbox directory per row, please.\n";
	    }
	    $new_mailto{$_} = 1;
	}

	unless( $err ) {

	    # remove old rows
	    foreach( keys %old_mailto_has ) {
		next if $new_mailto{$_} or /^s+$/s;
		push @removed, $self->{'target'} = $_;
		$self->master_execute('mailaddressremove');
		delete $cur_mailto{$_};
	    }
	    $self->{'status'} .= 
		" Removed @removed from $mail\@$self->{'domain'}\n"
		    if @removed;
	
	    # insert new rows
	    foreach( keys %new_mailto ) {
		next if $old_mailto_has{$_};
		push @added, $self->{'target'} = $_;
		$self->master_execute('mailaddressadd');
		$cur_mailto{$_} = 1;
	    }
	    $self->{'status'} .= 
		" Added @added into $mail\@$self->{'domain'}\n"
		    if @added;
	}
	$self->{'status'} = $err if $err;
    } # @$mailto

    my @data = keys %cur_mailto;
    foreach( @data ) { s/^&//; }
    $self->{'mailto'} = join("\n", sort @data) . "\n";

    $self->{'remove_button'} =
	'<INPUT TYPE=SUBMIT NAME=do_remove_mail VALUE="Remove Address" />';
}

sub page_new_mail {} # dummy

sub page_remove_mail {
    my $self = shift;
    $self->{'domain'} = $self->cleanparam('domain','[\w.-]{6,99}')
	or $self->page('menu','Missing domain name');
    $self->{'mail'} = $self->cleanparam('mail','(?:[\w.-]{1,99}|NEW)');
    $self->master_execute('mailaddresswipe');
    $self->page('domain',"Mail address '$self->{'mail'}' removed");
}

1;
